import pandas as pd
from nltk.stem.porter import PorterStemmer
import re


class NegativeRule(object):
    """Negative rules"""

    def __init__(self, left, right, id_column):
        self.left = self._preprocess(left, id_column)
        self.right = self._preprocess(right, id_column)
        self.id_column = id_column
        self.negative_rules = set()

    def get_tokens_diff(self, l_tokens, r_tokens):
        # get difference of l_token set and r_token set
        l_diff = [l - r for l, r in zip(l_tokens, r_tokens)]
        r_diff = [r - l for l, r in zip(l_tokens, r_tokens)]
        return l_diff, r_diff

    def _preprocess(self, df, id_column):
        """ Preprocess the records: (1) concatenate all columns. (2) lowercase,
        remove punctuation and do stemming (3) split by space

        Parameters
        ----------
        df: pd.DataFrame
            Original table

        id_column: string
            The name of id column in two tables.

        Reutrn
        ------
        result: pd.DataFrame
            Preprocessed table that has two columns, am id column named as "id"
            and a column for preprocessed record named "value"
        """
        # get column names except id
        columns = [c for c in df.columns if c != id_column]
        ps = PorterStemmer()

        # concat all columns, lowercase, remove punctuation, split by space,
        # and do stemming
        new_value = []
        for x in df[columns].values:
            concat_x = " ".join([str(i) for i in x])
            lower_x = re.sub('[^\w\s]', " ", concat_x.lower())
            stem_x = [ps.stem(w) for w in lower_x.split()]
            new_x = set(stem_x)
            new_value.append(new_x)

        id_df = df[id_column].values
        result = pd.DataFrame({id_column: id_df, "value": new_value})
        return result

    def learn(self, LL_blocked):
        """Learn opposite rules from LL"""
        # merge LL with left
        LL = self._merge(self.left, self.left, LL_blocked)

        # get token difference
        l_diff, r_diff = self.get_tokens_diff(LL["value_l"].values,
                                              LL["value_r"].values)

        # get rules: (l_token, r_token) that have one different token from each other
        for l, r, l_set in zip(l_diff, r_diff, LL["value_l"]):
            if len(l) == 1 and len(r) == 1 and len(l_set) != 1:
                self.negative_rules.add((list(l)[0], list(r)[0]))
                self.negative_rules.add((list(r)[0], list(l)[0]))

        # print(self.negative_rules)
        # raise

    def _merge(self, left, right, LR_blocked):
        id_column = self.id_column
        LR = LR_blocked[[id_column + "_l", id_column + "_r"]]
        LR = LR.merge(left, left_on=id_column + "_l", right_on=id_column)\
               .drop(columns=id_column) \
               .merge(right, left_on=id_column + "_r", right_on=id_column,
                   suffixes=("_l", "_r"))\
               .drop(columns=id_column)
        return LR

    def apply(self, LR_blocked):
        """Apply opposite rule on LR blocked"""
        # merge LR with left, right
        LR = self._merge(self.left, self.right, LR_blocked)

        # get token difference
        l_diff, r_diff = self.get_tokens_diff(LR["value_l"].values,
                                              LR["value_r"].values)

        # apply rule
        mask = []
        for lid, rid, l_d, r_d in zip(LR["autofj_id_l"].values,
                                      LR["autofj_id_r"].values,
                                      l_diff,
                                      r_diff):
            pairs = [(l, r) for l in l_d for r in r_d]
            meet_rule = any([p in self.negative_rules for p in pairs])
            mask.append(not meet_rule)

        LR_blocked = LR[mask][["autofj_id_l", "autofj_id_r"]]
        return LR_blocked



import pandas as pd
from ..utils import print_log
from multiprocessing import Pool
from functools import partial
from .join_function.autofj_join_function import AutoFJJoinFunction
from .options import autofj_lg, autofj_md, autofj_sm
import os
import shutil

class AutoFJJoinFunctionSpace(object):
    """AutoFJ Configuration Space. The space is specified by the space of join
    functions, the space of distance thresholds and the space of column weights.

    Parameters
    ----------
    join_function_space: string or dict or list, default="autofj_sm"
        There are following three ways to define the space of join functions:
        (1) Use the name of built-in join function space. There are three
        options, including "autofj_lg", "autofj_lg" and "autofj_sm" that use
        136, 68 and 14 join functions, respectively. Using less join functions
        can improve efficiency but may worsen performance.
        (2) Use a dict specifying the options for preprocessing methods,
        tokenization methods, token weighting methods and distance functions.
        The space will be the cartesian product of all options in the dict.
        See ./join_function_space/options.py for defining join functions using
        a dict.
        (3) Use a list of customized JoinFunction objects.

    n_jobs : int, default=1
        Number of CPU cores used. -1 means using all processors.

    verbose: bool, default=False
        Whether to print logging

    cache_dir: string, default="autofj_temp"
        Directory for cache
    """

    def __init__(self,
                 join_function_space="autofj_sm",
                 n_jobs=-1,
                 verbose=False,
                 cache_dir="autofj_temp"):
        if type(join_function_space) == str:
            builtin_jfs = {
                "autofj_sm": autofj_sm,
                "autofj_md": autofj_md,
                "autofj_lg": autofj_lg,
            }

            if join_function_space not in builtin_jfs:
                raise Exception("{} is not a valid join function space"
                                .format(join_function_space))

            self.join_functions = self._get_autofj_join_functions(
                builtin_jfs[join_function_space])
            self.use_builtin_fj_space = True
        elif type(join_function_space) == dict:
            self.join_functions = self._get_autofj_join_functions(
                join_function_space)
            self.use_builtin_fj_space = True
        else:
            self.join_functions = join_function_space
            self.use_builtin_fj_space = False

        self.verbose = verbose
        self.n_jobs = n_jobs if n_jobs > 0 else os.cpu_count()
        self.cache_dir = cache_dir

    def compute_distance(self, left, right, LL_blocked, LR_blocked):
        """Compute distance between each record pair in the blocked table

        Parameters
        ----------
        left: pd.DataFrame
            The left table. The id column is named as autofj_id.

        right: pd.DataFrame
            The right table. The id column is named as autofj_id.

        LL_blocked: pd.DataFrame
            The blocked table of left-left self-join, i.e., candidate set.
            The id columns are named as autofj_id_l and autofj_id_r,
            corresponding to the id column of left and another left table,
            respectively.

        LR_blocked: pd.DataFrame
            The blocked table of left-right join, i.e., candidate set.
            The id columns are named as autofj_id_l and autofj_id_r,
            corresponding to the id column of left and right table, respectively.

        Return:
        -------
        LR_distance: dict {string: pd.DataFrame}
            A dict of distance tables measured by different join functions. The
            key of dict is the name of the join function. The value is a
            distance table that contains distance between tuple pairs for each
            column (see the example below).

            Example of distance table:
            | autofj_id_l | autofj_id_r | column 1 | column 2 | column 3 |
            |------------------------------------------------------------|
            |      0      |      1      |    0.5   |     0.2  |     1    |
            |      0      |      2      |    0.3   |     0.4  |    0.7   |
        """
        # sort by ids
        LL_blocked = LL_blocked[["autofj_id_l", "autofj_id_r"]] \
            .sort_values(by=["autofj_id_r", "autofj_id_l"]) \
            .reset_index(drop=True)
        LR_blocked = LR_blocked[["autofj_id_l", "autofj_id_r"]] \
            .sort_values(by=["autofj_id_r", "autofj_id_l"]) \
            .reset_index(drop=True)

        # build cache for computing distance
        if self.use_builtin_fj_space:
            self.build_cache(left, right, self.cache_dir)

        # compute distance for each column
        column_names = [c for c in left.columns if c != "autofj_id"]
        args = []
        for c in column_names:
            for jf in self.join_functions:
                args.append((c, jf))

        func = partial(self._compute_column_distance, left=left, right=right,
                       LL_blocked=LL_blocked, LR_blocked=LR_blocked,
                       cache_dir=self.cache_dir)

        if self.n_jobs == 1:
            column_distances = [func(arg) for arg in args]
        else:
            with Pool(self.n_jobs) as pool:
                column_distances = pool.map(func, args)

        # remove cache
        if self.use_builtin_fj_space:
            self.remove_cache(self.cache_dir)

        # get distance table for each join function
        LL_distance = {}
        LR_distance = {}

        for i, (c, jf) in enumerate(args):
            jf_name = jf.name
            if jf_name not in LL_distance:
                LL_distance[jf_name] = {
                    "autofj_id_l": LL_blocked["autofj_id_l"].values,
                    "autofj_id_r": LL_blocked["autofj_id_r"].values}
                LR_distance[jf_name] = {
                    "autofj_id_l": LR_blocked["autofj_id_l"].values,
                    "autofj_id_r": LR_blocked["autofj_id_r"].values}

            LL_d, LR_d = column_distances[i]
            LL_distance[jf_name][c] = LL_d
            LR_distance[jf_name][c] = LR_d

        for jf_name in LL_distance.keys():
            LL_distance[jf_name] = pd.DataFrame(LL_distance[jf_name])
            LR_distance[jf_name] = pd.DataFrame(LR_distance[jf_name])

        return LL_distance, LR_distance

    def _compute_column_distance(self, arg, left, right, LL_blocked, LR_blocked,
                                cache_dir):
        """Compute distance for a column using a join function


        Parameters
        ----------
        arg: tuple
            A tuple that consists of column_name and join function object

        Returns
        -------
        LL_d: pd.Series
            Distance of tuple pairs in LL_blocked

        LR_d: pd.Series
            Distance of tuple pairs in LR_blocked
        """
        c, jf = arg
        if self.verbose:
            print_log("{}: Compute distances for column {}.".format(jf.name, c))

        # get subsets of column c from tables
        l = left[["autofj_id", c]].rename(columns={c: "value"})
        r = right[["autofj_id", c]].rename(columns={c: "value"})

        cache_dir = os.path.join(cache_dir, c)

        # compute distances for column c
        LL_d, LR_d = jf.compute_distance(l, r, LL_blocked, LR_blocked, cache_dir)
        return LL_d, LR_d

    def build_cache(self, left, right, cache_dir="autofj_temp"):
        """ Build cache for computing distance

        Parameters
        ----------
        left: pd.DataFrame
            The left table. The id column is named as autofj_id.

        right: pd.DataFrame
            The right table. The id column is named as autofj_id.

        cache_dir: string
            Directory to save cache

        """
        # indexing
        left = left.set_index("autofj_id")
        right = right.set_index("autofj_id")

        # build cache for each column of left and right table
        column_names = [c for c in left.columns if c != "autofj_id"]
        func = partial(self._build_cache_column, left=left, right=right,
                       cache_dir=cache_dir)
        args = []
        for c in column_names:
            args.append((c, "left"))
            args.append((c, "right"))

        with Pool(self.n_jobs) as pool:
            pool.map(func, args)

    def _build_cache_column(self, column_table, left, right, cache_dir):
        """Built cache on one column

        Parameters
        ----------
        column: string
            The column to build cache
        """
        # get data under processing
        column, table = column_table
        if table == "left":
            X = left[column]
        else:
            X = right[column]

        preprocess_results = {}
        tokenize_results = {}
        token_weights = {}
        column_cache_dir = os.path.join(cache_dir, column)

        # cache results
        for jf in self.join_functions:
            # get cache paths
            pre_path, pre_token_path, weight_path \
                = jf.get_cache_paths(column_cache_dir, table)
            p = jf.preprocess_method
            t = jf.tokenize_method
            w = jf.token_weight_method

            if p not in preprocess_results:
                X_pre = jf.preprocess(X, cache_path=pre_path)
                preprocess_results[p] = X_pre
            else:
                X_pre = preprocess_results[p]

            if (p, t) not in tokenize_results:
                X_pre_token = jf.tokenize(X_pre, cache_path=pre_token_path)
                tokenize_results[(p, t)] = X_pre_token
            else:
                X_pre_token = tokenize_results[(p, t)]

            if table == "left" and (p, t, w) not in token_weights:
                weights = jf.compute_token_weights(X_pre_token,
                                                   cache_path=weight_path)
                token_weights[(p, t, w)] = weights

    def remove_cache(self, cache_dir="autofj_temp"):
        if os.path.exists(cache_dir):
            shutil.rmtree(cache_dir)

    def _get_autofj_join_functions(self, options):
        """ Get a space of built-in join functions with different options. The
        space will be formed by making cartesian product on all different
        options.

        Parameters
        ----------
        options: dict
            The options of preprocess methods, tokenization methods, token weights
            and distance functions. See option.py for examples. The dict contains
            the following key-values:
            - preprocess_methods: list of string
                A list of preprocess methods defined in preprocessor.py
            - tokenize_methods: list of string
                A list of tokenization methods defined in tokenizer.py
            - token_weights: list of string
                A list of tokenization methods defined in token_weights.py
            - char_distance_functions: list of string
                A list of character-based distance functions defined in
                distance_function.py
            - set_distance_functions: list of string
                A list of set-based distance functions defined in
                distance_function.py

        Return
        ------
        join_functions: list of AutoFJJoinFunction object
            A list of join functions.
        """
        join_functions = []
        # join functions using set-based distance functions
        for p in options["preprocess_methods"]:
            for t in options["tokenize_methods"]:
                for w in options["token_weights"]:
                    for d in options["set_distance_functions"]:
                        jf = AutoFJJoinFunction(p, t, w, d)
                        join_functions.append(jf)

        # join functions using char-based distance functions
        for p in options["preprocess_methods"]:
            for d in options["char_distance_functions"]:
                jf = AutoFJJoinFunction(p, None, None, d)
                join_functions.append(jf)
        return join_functions


import pandas as pd
import math
import collections
import ngram
import os
from multiprocessing import Pool
import re
from nltk.stem.porter import PorterStemmer
import time
from functools import partial
import numpy as np

class AutoFJBlocker(object):
    """AutoFJ default Blocker

    Parameters
    ----------
    num_candidates: int or None
        Number of candidates for each right record. If None, the number will be
        the square root of the number of records in the left table

    n_jobs : int, default=1
        Number of CPU cores used. -1 means using all processors.
    """
    def __init__(self, num_candidates=None, n_jobs=-1):
        self.n_jobs = n_jobs if n_jobs > 0 else os.cpu_count()
        self.num_candidates = num_candidates

    def block(self, left_table, right_table, id_column):
        """ Perform blocking on two tables

        Parameters
        ----------
        left_table: pd.DataFrame
            Reference table. The left table is assumed to be almost
            duplicate-free, which means it has no or only few duplicates.

        right_table: pd.DataFrame
            Another input table.

        id_column: string
            The name of id column in two tables.

        Returns:
        --------
            result: pd.DataFrame
            A table of records pairs survived blocking. Column names
            id_column + "_l" and id_column + "_r"
        """
        self.id_column = id_column

        # preprocess records
        left = self._preprocess(left_table, id_column)
        right = self._preprocess(right_table, id_column)

        # get num candidates
        if self.num_candidates is None:
            self.num_candidates = min(int(math.pow(len(left), 0.5)), 50)

        # build token maps using the left table
        token_lid_map, token_idf_map, token_tf_map = self._build_token_maps(
            left)

        # get candidates for each right record
        result = self._get_candidates_multi(right, token_lid_map,
                                      token_idf_map, token_tf_map)

        result = result.rename(columns={"id_l": id_column+"_l",
                                        "id_r": id_column + "_r"})
        return result

    def _preprocess(self, df, id_column):
        """ Preprocess the records: (1) concatenate all columns. (2) lowercase,
        remove punctuation and do stemming

        Parameters
        ----------
        df: pd.DataFrame
            Original table

        id_column: string
            The name of id column in two tables.

        Reutrn
        ------
        result: pd.DataFrame
            Preprocessed table that has two columns, am id column named as "id"
            and a column for preprocessed record named "value"
        """
        # get column names except id
        columns = [c for c in df.columns if c != id_column]
        ps = PorterStemmer()

        # concat all columns, lowercase, remove punctuation, split by space,
        # and do stemming
        new_value = []
        for x in df[columns].values:
            concat_x = " ".join([str(i) for i in x])
            lower_x = re.sub('[^\w\s]', " ", concat_x.lower())
            stem_x = [ps.stem(w) for w in lower_x.split()]
            new_x = " ".join(stem_x)
            new_value.append(new_x)

        id_df = df[id_column].values
        result = pd.DataFrame({"id":id_df, "value":new_value})
        return result

    def _build_token_maps(self, left):
        """ Build token maps for blocking

        Parameters:
        -----------
        left: pd.DataFrame
            The left table

        Return:
        token_id_map: dict {token: left_id_set]}
            A dictionary that maps each token to id of left records contain
            the token

        token_idf_map: dict {token: idf_score}
            A dictionary that maps each token to its idf score

        token_tf_map: dict {token: {left_id: tf_score}}
            A dictionary that maps each token to its tf scores in different
            left records.
        """
        token_lid_map = collections.defaultdict(set)
        token_tf_map = collections.defaultdict(dict)
        three_gram = ngram.NGram(N=3)

        # build token maps
        for lid, lvalue in left[["id", "value"]].values:
            if len(lvalue) == 0:
                continue

            # 3-gram tokenization + white space tokenization
            tokens = list(three_gram.split(lvalue)) + list(lvalue.split())

            # add lid to id map
            for token in tokens:
                token_lid_map[token].add(lid)

            #compute tf score
            counter = collections.Counter(tokens)

            for token, count in counter.items():
                token_tf_map[lid][token] = count / len(tokens)

        # compute idf score
        token_idf_map = {}
        for token, lids in token_lid_map.items():
            token_idf_map[token] = math.log(len(left) / (len(lids)+1))
        return token_lid_map, token_idf_map, token_tf_map

    def _get_candidates(self, right, token_lid_map, token_idf_map,
                            token_tf_map):
        """ Get candidates for one record in right table
        Parameters:
        -----------
        right: pd.DataFrame
            Right table

        token_id_map: dict {token: left_id_set]}
            A dictionary that maps each token to id of left records contain
            the token

        token_idf_map: dict {token: idf_score}
            A dictionary that maps each token to its idf score

        token_tf_map: dict {token: {left_id: tf_score}}
            A dictionary that maps each token to its tf scores in different
            left records.

        Return:
        -------
        result: pd.DataFrame
            A table with two columns "id_l", "id_r" that are ids of candidate
            record pairs
        """
        three_gram = ngram.NGram(N=3)
        result = []
        
        for rid, rvalue in right[["id", "value"]].values:
            tokens = set(list(three_gram.split(rvalue)) + list(rvalue.split()))
            counter = collections.defaultdict(int)

            for token in tokens:
                for lid in token_lid_map[token]:
                    counter[lid] += token_idf_map[token] *\
                                    token_tf_map[lid][token]

            if len(counter) <= self.num_candidates:
                candidate_lids = list(counter.keys())
            else:
                counter_sorted = sorted(counter, key=counter.get)[::-1]
                thresh = counter[counter_sorted[self.num_candidates-1]]
                candidate_lids = [lid for lid, value in counter.items()
                                  if value >= thresh]

            for lid in candidate_lids:
                result.append([lid, rid])

        result = pd.DataFrame(result, columns=["id_l", "id_r"])
        return result

    def _get_candidates_multi(self, right, token_lid_map, token_idf_map,
                            token_tf_map):
        """ Get candidates for one record in right table using multiple cpus
        Parameters:
        -----------
        right: pd.DataFrame
            Right table

        token_id_map: dict {token: left_id_set]}
            A dictionary that maps each token to id of left records contain
            the token

        token_idf_map: dict {token: idf_score}
            A dictionary that maps each token to its idf score

        token_tf_map: dict {token: {left_id: tf_score}}
            A dictionary that maps each token to its tf scores in different
            left records.

        Return:
        -------
        result: pd.DataFrame
            A table with two columns "id_l", "id_r" that are ids of candidate
            record pairs
        """
        right_groups = np.array_split(right, self.n_jobs)
        func = partial(self._get_candidates, token_lid_map=token_lid_map,
                       token_idf_map=token_idf_map, token_tf_map=token_tf_map)

        with Pool(self.n_jobs) as pool:
            results = pool.map(func, right_groups)

        results = pd.concat(results, axis=0).sort_values(by=["id_r", "id_l"])
        return  results

if __name__ == '__main__':
    # Test Case
    tic = time.time()
    blocker = AutoFJBlocker(n_jobs=4)
    left = pd.read_csv("data/left.csv")
    right = pd.read_csv("data/right.csv")
    result = blocker.block(left, right, 'id')
    print(result)
    print(time.time() - tic)


class JoinFunction(object):
    """Customized join function must have an unique name attribute and a method named 
       compute_distance as below. The constructor can be overwritten"""
    def __init__(self):
        self.name = "jf_example"
        pass

    def compute_distance(self, left, right, LL_blocked, LR_blocked,
                         cache_dir=None):
        """Compute the distance of each tuple pair in the LL and LR blocked table.

        Parameters
        ----------
        left: pd.DataFrame
            A subset of the left table that contains the id column and the
            column to be processed. The id column is named as autofj_id.
            The column to be processed is named as value.

        right: pd.DataFrame
            A subset of the right table that contains the id column and the
            column to be processed. The id column is named as autofj_id.
            The column to be processed is named as value.

        LL_blocked: pd.DataFrame
            The LL blocked table that consists of the id columns and
            the columns to be processed. The id columns are named as
            autofj_id_l and autofj_id_r. The column to be processed is named as
            value_l and value_r.

        LR_blocked: pd.DataFrame
            The LR blocked table that consists of the id columns and
            the columns to be processed. The id columns are named as
            autofj_id_l and autofj_id_r. The column to be processed is named as
            value_l and value_r.

        Returns
        -------
        LL_distance: pd.Series
            Distance of each tuple pair in the LL blocked table.

        LR_distance: pd.Series
            Distance of each tuple pair in the LR blocked table.
        """
        LL_distance = None
        LR_distance = None
        return LL_distance, LR_distance



class Blocker(object):
    """The customized blocker must have a block method as below. The constructor
    can be overwritten"""
    def __init__(self):
        pass

    def block(self, left_table, right_table, id_column):
        """ Perform blocking on two tables

        Parameters
        ----------
        left_table: pd.DataFrame
            Reference table. The left table is assumed to be almost
            duplicate-free, which means it has no or only few duplicates.

        right_table: pd.DataFrame
            Another input table.

        id_column: string
            The name of id column in two tables.

        Returns:
        --------
            result: pd.DataFrame
            A table of records pairs survived blocking. Column names
            id_column + "_l" and id_column + "_r"
        """
        result = None
        return result


from datetime import datetime
import os

def print_log(message):
    print("{}:{}".format(datetime.now().strftime('%H:%M:%S'), message))

def makedir(dir_list, file=None, remove_old_dir=False):
    save_dir = os.path.join(*dir_list)

    if remove_old_dir and os.path.exists(save_dir) and file is None:
        shutil.rmtree(save_dir)

    if not os.path.exists(save_dir):
        os.makedirs(save_dir)
    if file is not None:
        save_dir = os.path.join(save_dir, file)
    return save_dir


import numpy as np
import time
import os
from copy import deepcopy
from ..utils import print_log


class AutoFJGreedyAlgorithm(object):
    """Greedy Algorithm for single column datasets. Select a union of join
    configurations that maximize the recall while keeping the precision above
    the precision target.

    Parameters
    ----------
    LL_distance: dict
        Distance of tuple pairs in LL tables on one column measured by different
        join functions. The distance is saved in a dict. The key is the name
        of join functions. The value is a table of distance between tuple pairs
        on one column. The first two columns in the table are "autofj_id_l",
        "autofj_id_r". The third column is the "distance".

    LR_distance: dict
        Distance of tuple pairs in LR tables on one column measured by different
        join functions. The distance is saved in a dict. The key is the name
        of join functions. The value is a table of distance between tuple pairs
        on one column. The first two columns in the table are "autofj_id_l",
        "autofj_id_r". The third column is the "distance"

    precision_target: float
        Precision target.

    unscaled_candidate_thresholds: list
        Candidates of thresholds

    n_jobs : int, default=1
        Number of CPU cores used. -1 means using all processors.
    """

    def __init__(self,
                 LL_distance,
                 LR_distance,
                 precision_target,
                 unscaled_candidate_thresholds,
                 patience=0.05,
                 verbose=False,
                 n_jobs=-1):
        self.target = precision_target
        self.full_LR_distance = LR_distance
        self.LL_distance = self.groupby_rid(LL_distance)
        self.LR_distance = self.keep_min_distance(LR_distance)
        self.join_functions = sorted(list(LL_distance.keys()))
        self.candidate_thresholds = self.scale_candidate_thresholds(
            unscaled_candidate_thresholds)
        self.n_jobs = n_jobs if n_jobs > 0 else os.cpu_count()
        self.precision_cache = self.precompute_precision()
        self.patience = patience
        self.verbose=verbose

    def keep_min_distance(self, LR_distance):
        # for each rid only keep lid with the smallest distance
        LR_small = {}
        for config, LR in LR_distance.items():
            LR_s = LR.sort_values(by=["distance", "autofj_id_l"],
                                  ascending=True).drop_duplicates(
                subset="autofj_id_r", keep="first")
            LR_s = LR_s.sort_values(by=["autofj_id_r", "autofj_id_l"])
            LR_small[config] = LR_s
        return LR_small

    def groupby_rid(self, LL_distance):
        """Group the distance table by rid. lids and distance for one rid will
         be put in one table.
         """
        LL_group = {}
        for config, LL in LL_distance.items():
            rids = LL["autofj_id_r"].values
            break_indices = np.argwhere(np.diff(rids) > 0).ravel()
            values = np.split(LL, break_indices + 1)
            keys = [rids[0]] + rids[break_indices + 1].tolist()
            LL_dict = {keys[i]: values[i] for i in range(len(values))}
            LL_group[config] = LL_dict
        return LL_group

    def scale_candidate_thresholds(self, unscaled_thresholds):
        """Scale candidate thresholds for each join function such that the
           threshold is in the range of [min, max] of LR distance.
        """
        candidate_thresholds = {}
        unscaled_thresholds = np.array(unscaled_thresholds)

        for jf in self.join_functions:
            max_d = self.LR_distance[jf]["distance"].values.max()
            min_d = self.LR_distance[jf]["distance"].values.min()
            cand_thresh = unscaled_thresholds * (max_d - min_d) + min_d
            candidate_thresholds[jf] = set(cand_thresh.tolist())
        return candidate_thresholds

    def precompute_precision(self):
        """ Precompute precision of each LR join for every join configuration
        The precision is computed as 1/ #L-L joins
        Returns
        -------
        precision: dict
            The precision of each LR join for every join configuration. The key
            is the join configuration (join function, threshold). The value is
            a dict, where the key is rid and value is a tuple (lid, precision).
        """
        precision = {}
        for jf in self.join_functions:
            LL = self.LL_distance[jf]
            LR = self.LR_distance[jf]
            for thresh in self.candidate_thresholds[jf]:
                prec = {}

                for lid, rid, d in LR.values:
                    if d > thresh:
                        continue
                    # compute precision as 1/ #L-L joins
                    num_LL_joins = self.get_num_LL_joins(LL, lid, thresh)
                    prec[rid] = (lid, 1 / num_LL_joins)

                precision[(jf, thresh)] = prec

        return precision

    def get_num_LL_joins(self, LL, proxy, thresh):
        """ Number of L-L joins of proxy (the l record closest to each R),
        which is the number of L records that have distance smaller than
        2 * threshold to the proxy"""
        if proxy not in LL:
            return 1
        else:
            lid_df = LL[proxy]
            mask = lid_df["distance"] <= 2 * thresh
            num_proxy_joins = mask.sum() + 1
            return num_proxy_joins

    def compute_profit(self, added_config):
        """ Compute profit of adding one join configuration to the union. The
        profit is defined as true positive (TP) / false positive (FP).

        Parameters
        ----------
        added_config: tuple
            Join configuration to be added (join function, threshold)

        Returns
        -------
        profit: float
            Estimated TP / Estimated FP
        """
        # extract precision from cache
        precision = self.precision_cache[added_config]

        # compute the change of TP and number of joins by adding the new config
        TP = self.running_TP
        n_joins = self.running_n_joins

        for r, (l, p) in precision.items():
            if r in self.running_l_cands:
                # If r has been joined before, conflict may occur.
                # To resolve conflicts, use the join with the highest precision.
                # Only this join is counted toward TP. But other joins are
                # preserved in FP to penalize on conflicts.
                l_cands = self.running_l_cands[r]
                old_p = self.running_local_prec[r]
                delta_TP = max(old_p, p) - old_p
                delta_n_joins = 0 if l in l_cands else 1
            else:
                delta_TP = p
                delta_n_joins = 1

            TP += delta_TP
            n_joins += delta_n_joins

        FP = n_joins - TP
        avg_prec = TP / (n_joins + 1e-9)

        if n_joins == self.running_n_joins:
            # ensure some new records are joined
            profit = float("-inf")
        else:
            profit = TP / (FP + 1e-9)

        return profit, avg_prec

    def update_selection(self, config, avg_prec):
        """Updated selected join configuration"""
        self.running_configs.append(config)

        for r, (l, p) in self.precision_cache[config].items():
            if r in self.running_l_cands:
                l_cands = self.running_l_cands[r]
                old_p = self.running_local_prec[r]

                if p > old_p:
                    self.running_local_prec[r] = p
                    self.running_LR_joins[r] = l
                    delta_TP = p - old_p
                else:
                    delta_TP = 0

                if l in l_cands:
                    delta_n_joins = 0
                else:
                    self.running_l_cands[r].add(l)
                    delta_n_joins = 1
            else:
                self.running_l_cands[r] = {l}
                self.running_local_prec[r] = p
                self.running_LR_joins[r] = l
                delta_TP = p
                delta_n_joins = 1

            self.running_TP += delta_TP
            self.running_n_joins += delta_n_joins

        if avg_prec > self.target:
            self.global_prec = avg_prec
            self.config_selected = deepcopy(self.running_configs)
            self.LR_joins = [(l, r) for r, l in self.running_LR_joins.items()]

    def run(self):
        """Run greedy algorithm"""
        self.running_l_cands = {}
        self.running_local_prec = {}
        self.running_configs = []
        self.running_LR_joins = {}
        self.running_n_joins = 0
        self.running_TP = 0

        self.config_selected = None
        self.LR_joins = None
        self.global_prec = None

        candidate_configs = []
        for jf in self.join_functions:
            for thresh in self.candidate_thresholds[jf]:
                candidate_configs.append((jf, thresh))

        n_iter = 1
        while len(candidate_configs) > 0:
            best_profit = float("-inf")
            best_config = None
            best_avg_prec = None

            for i, config in enumerate(candidate_configs):
                profit, avg_prec = self.compute_profit(config)

                if profit > best_profit:
                    best_profit = profit
                    best_config = config
                    best_avg_prec = avg_prec

            if best_config is None:
                break

            self.update_selection(best_config, best_avg_prec)

            if self.verbose:
                print_log("Iteration {}, precision {}, # LR joins {}".\
                                format(n_iter,
                                       best_avg_prec,
                                       len(self.running_LR_joins)))

            if best_avg_prec < self.target - self.patience:
                if self.verbose:
                    print_log("Exit because precision below target")
                break

            n_iter += 1

        return self.LR_joins, self.config_selected

    def get_full_LR_sel(self):
        """Get full LR matches instead of selecting one l for each r
           Note: Only work for processing one config at each time.
        """
        if self.config_selected is None:
            return None

        config, p = self.config_selected[0]
        full_LR = self.full_LR_distance[config]
        mask = full_LR["distance"].values <= p
        lr_sel = [(l, r) for l, r in
                  full_LR[mask][["autofj_id_l", "autofj_id_r"]].values]
        return lr_sel

    def get_reward(self):
        if self.config_selected is None:
            return float("-inf")

        return self.global_prec * len(self.LR_joins)

    def get_config_selected(self):
        return self.config_selected


import ngram
import pandas as pd
import numpy as np
import time

three_gramer = ngram.NGram(N=3)

def splitBySpace(x):
    return x.split()

def threeGram(x):
    # Replace whitespace more than one by a single blank
    return list(three_gramer.split(" ".join(x.split())))

class Tokenizer:
    """Tokenize data

    Parameters
    ----------
    method: string
        Tokenization method. The available methods are listed as follows.
        - splitBySpace
        - threeGram
        - None (no tokenization)
    """
    def __init__(self, method):
        self.method = method
        if method is None:
            self.func = None
        elif method == "splitBySpace":
            self.func = splitBySpace
        elif method == "threeGram":
            self.func = threeGram
        else:
            raise Exception("{} is an invalid tokenization method"
                             .format(method))

    def tokenize(self, X):
        """ Tokenize input data

        Parameters
        ----------
        X: pd.Series
            Input data
        """
        if self.func is not None:
            X = X.apply(self.func)
        return X

# data = pd.read_csv("../../data/left.csv")["title"]
# X = np.concatenate([data.values for _ in range(20)])
# X = pd.Series(X)
#
# tokenizer = Tokenizer("threeGram")
# tic = time.time()
# tokenizer.tokenize(X)
# print(time.time() - tic)


from .join_function_space.autofj_join_function_space import AutoFJJoinFunctionSpace
from .blocker.autofj_blocker import AutoFJBlocker
from .optimizer.autofj_multi_column_greedy_algorithm import \
    AutoFJMulticolGreedyAlgorithm
import pandas as pd
from .utils import print_log
import os
from .negative_rule import NegativeRule
import numpy as np


class AutoFJ(object):
    """
    AutoFJ automatically produces record pairs that approximately match in 
    two tables L and R. It proceeds to configure suitable parameters 
    automatically, which when used to fuzzy-join L and R, meets the 
    user-specified precision target, while maximizing recall.
    
    AutoFJ attempts to solve many-to-one join problems, where each record in R
    will be joined with at most one record in L, but each record in L can be 
    joined with multiple records in R. In AutoFJ, L refers to a reference 
    table, which is assumed to be almost "duplicate-free".

    Parameters
    ----------
    precision_target: float, default=0.9
        Precision target.

    join_function_space: string or dict or list of objects, default="autofj_sm"
        There are following three ways to define the space of join functions:
        (1) Use the name of built-in join function space. There are three
        options, including "autofj_lg", "autofj_lg" and "autofj_sm" that use
        136, 68 and 14 join functions, respectively. Using less join functions
        can improve efficiency but may worsen performance.
        (2) Use a dict specifying the options for preprocessing methods,
        tokenization methods, token weighting methods and distance functions.
        The space will be the cartesian product of all options in the dict.
        See ./join_function_space/options.py for defining join functions using
        a dict.
        (3) Use a list of customized JoinFunction objects.

    distance_threshold_space: int or list, default=50
        The number of candidate distance thresholds or a list of candidate
        distance thresholds in the space.  If the number of distance thresholds
        (integer) is given, distance thresholds are spaced evenly from 0 to 1.
        Otherwise, it should be a list of floats from 0 to 1.

    column_weight_space: int or list, default=10
        The number of candidate column weights or a list of candidate
        column weights in the space. If the number of column weights
        (integer) is given, column weights are spaced evenly from 0 to 1.
        Otherwise, it should be a list of floats from 0 to 1.

    blocker: a Blocker object or None, default None
        A Blocker object that performs blocking on two tables. If None, use 
        the built-in blocker. For customized blocker, see Blocker class.

    n_jobs : int, default=-1
        Number of CPU cores used. -1 means using all processors.

    verbose: bool, default=False
        Whether to print logging
    """

    def __init__(self,
                 precision_target=0.9,
                 join_function_space="autofj_sm",
                 distance_threshold_space=50,
                 column_weight_space=10,
                 blocker=None,
                 n_jobs=-1,
                 verbose=False):
        self.precision_target = precision_target
        self.join_function_space = join_function_space

        if type(distance_threshold_space) == int:
            self.distance_threshold_space = list(
                np.linspace(0, 1, distance_threshold_space))
        else:
            self.distance_threshold_space = distance_threshold_space

        if type(column_weight_space) == int:
            self.column_weight_space = list(
                np.linspace(0, 1, column_weight_space))
        else:
            self.column_weight_space = column_weight_space

        if blocker is None:
            self.blocker = AutoFJBlocker(n_jobs=n_jobs)
        else:
            self.blocker = blocker

        self.n_jobs = n_jobs if n_jobs > 0 else os.cpu_count()
        self.verbose = verbose

    def join(self, left_table, right_table, id_column, on=None):
        """Join left table and right table.

        Parameters
        ----------
        left_table: pd.DataFrame
            Reference table. The left table is assumed to be almost
            duplicate-free, which means it has no or only few duplicates.

        right_table: pd.DataFrame
            Another input table.

        id_column: string
            The name of id column in the two tables. This column will not be
            used to join two tables.

        on: list or None
            A list of column names (multi-column fuzzy join) that the two tables
            will be joined on. If None, two tables will be joined on all columns
            that exist in both tables, excluding the id column.

        Returns:
        --------
            result: pd.DataFrame
                A table of joining pairs. The columns of left table are
                suffixed with "_l" and the columns of right table are suffixed
                with "_r"
        """
        left = left_table.copy(deep=True)
        right = right_table.copy(deep=True)

        # create internal id columns (use internal ids)
        left["autofj_id"] = range(len(left))
        right["autofj_id"] = range(len(right))

        # remove original ids
        left.drop(columns=id_column, inplace=True)
        right.drop(columns=id_column, inplace=True)

        # get names of columns to be joined
        if on is None:
            on = sorted(list(set(left.columns).intersection(right.columns)))
        left = left[on]
        right = right[on]

        # do blocking
        if self.verbose:
            print_log("Start blocking")
        LL_blocked = self.blocker.block(left, left, "autofj_id")
        LR_blocked = self.blocker.block(left, right, "autofj_id")

        # remove equi-joins on LL
        LL_blocked = LL_blocked[
            LL_blocked["autofj_id_l"] != LL_blocked["autofj_id_r"]]

        # learn and apply negative rules
        nr = NegativeRule(left, right, "autofj_id")
        nr.learn(LL_blocked)
        LR_blocked = nr.apply(LR_blocked)

        # create join function space
        jf_space = AutoFJJoinFunctionSpace(self.join_function_space,
                                           n_jobs=self.n_jobs)

        # compute distance
        if self.verbose:
            print_log("Start computing distances. Size of join function space: {}"
                      .format(len(jf_space.join_functions)))

        LL_distance, LR_distance = jf_space.compute_distance(left,
                                                             right,
                                                             LL_blocked,
                                                             LR_blocked)

        # run greedy algorithm
        if self.verbose:
            print_log("Start running greedy algorithm.")

        optimizer = AutoFJMulticolGreedyAlgorithm(
            LL_distance,
            LR_distance,
            precision_target=self.precision_target,
            candidate_thresholds=self.distance_threshold_space,
            candidate_column_weights=self.column_weight_space,
            n_jobs=self.n_jobs
        )

        self.selected_column_weights, self.selected_join_configs, LR_joins = \
            optimizer.run()

        if LR_joins is None:
            print("Warning: The precision target cannot be achieved.",
                  "Try a lower precision target or a larger space of join functions,",
                  "distance thresholds and column weights.")
            LR_joins = pd.DataFrame(columns=[c+"_l" for c in left_table.columns]+
                                            [c+"_r" for c in right_table.columns])
            return LR_joins

        # merge with original left and right tables
        left_idx = [l for l, r in LR_joins]
        right_idx = [r for l, r in LR_joins]
        L = left_table.iloc[left_idx].add_suffix("_l").reset_index(drop=True)
        R = right_table.iloc[right_idx].add_suffix("_r").reset_index(drop=True)
        result = pd.concat([L, R], axis=1).sort_values(by=id_column + "_r")
        return result



"""Compute distance"""
import editdistance
import jellyfish
import collections
from collections import Counter
import time
import numpy as np
import pandas as pd
import spacy


"""Distance Functions"""
def jaccardDistance(x, y, w=None):
    inter = set(x).intersection(set(y))
    union = set(x).union(set(y))
    if w is None:
        sum_inter = len(inter)
        sum_union = len(union)
    else:
        sum_inter = sum([w[s] for s in inter])
        sum_union = sum([w[s] for s in union])
    d = 1 - sum_inter / (sum_union + 1e-9)
    return d

def cosineDistance(x, y, w=None):
    c1 = Counter(x)
    c2 = Counter(y)
    inter = set(x).intersection(set(y))

    if w is None:
        uv = sum([c1[s]*c2[s] for s in inter])
        u = np.sqrt(sum([c1[s]**2 for s in set(x)]))
        v = np.sqrt(sum([c2[s]**2 for s in set(y)]))
    else:
        uv = sum([w[s]*c1[s]*w[s]*c2[s] for s in inter])
        u = np.sqrt(sum([(w[s]*c1[s])**2 for s in set(x)]))
        v = np.sqrt(sum([(w[s]*c2[s])**2 for s in set(y)]))

    d = 1 - uv / (u * v + 1e-9)
    return d

def diceDistance(x, y, w=None):
    inter = set(x).intersection(set(y))
    union = set(x).union(set(y))
    if w is None:
        sum_inter = len(inter)
        sum_union = len(union)
    else:
        sum_inter = sum([w[s] for s in inter])
        sum_union = sum([w[s] for s in union])
    d = 1 - (2 * sum_inter / (sum_inter + sum_union + 1e-9))
    return d

def maxincDistance(x, y, w=None):
    inter = set(x).intersection(set(y))
    if w is None:
        sum_inter = len(inter)
    else:
        sum_inter = sum([w[s] for s in inter])

    if w is None:
        sum_x = len(set(x))
        sum_y = len(set(y))
    else:
        sum_x = sum([w[s] for s in set(x)])
        sum_y = sum([w[s] for s in set(y)])
    min_sum = min(sum_x, sum_y)
    d = 1 - (sum_inter / (min_sum + 1e-9))
    return d

def intersectDistance(x, y, w=None):
    inter = set(x).intersection(set(y))
    union = set(x).union(set(y))
    if w is None:
        sum_inter = len(inter)
        sum_union = len(union)
    else:
        sum_inter = sum([w[s] for s in inter])
        sum_union = sum([w[s] for s in union])
    d = 1 - sum_inter / (sum_inter + sum_union + 1e-9)
    return d

def isContain(x, y):
    set_x = set(x)
    set_y = set(y)

    if len(set_x) > len(set_y):
        return set_y.issubset(set_x)
    else:
        return set_x.issubset(set_y)

def containCosineDistance(x, y, w=None):
    if isContain(x, y):
        return cosineDistance(x, y, w)
    else:
        return 1

def containJaccardDistance(x, y, w=None):
    if isContain(x, y):
        return jaccardDistance(x, y, w)
    else:
        return 1

def containDiceDistance(x, y, w=None):
    if isContain(x, y):
        return diceDistance(x, y, w)
    else:
        return 1

def editDistance(x, y):
    d = editdistance.eval(x, y)
    return d

def jaroDistance(x, y):
    d = 1 - jellyfish.jaro_winkler_similarity(x, y)
    return d

def embedDistance(x, y, embedding):
    x = embedding(x)
    y = embedding(y)
    d = 1 - x.similarity(y)
    return d

class DistanceFunction(object):
    """Distance function

    Parameters
    ----------
    method: string
        Method of computing distance. The available methods are listed as
        follows.
        Set-based distance
            - jaccardDistance
            - cosineDistance
            - diceDistance
            - maxincDistance
            - intersectDistance
            - containCosineDistance
            - containJaccardDistance
            - containDiceDistance
        Char-based distance
            - editDistance
            - jaroDistance

    """
    def __init__(self, method):
        self.method = method
        if method == "jaccardDistance":
            self.func = jaccardDistance
        elif method == "cosineDistance":
            self.func = cosineDistance
        elif method == "diceDistance":
            self.func = diceDistance
        elif method == "maxincDistance":
            self.func = maxincDistance
        elif method == "intersectDistance":
            self.func = intersectDistance
        elif method == "editDistance":
            self.func = editDistance
        elif method == "jaroDistance":
            self.func = jaroDistance
        elif method == "containCosineDistance":
            self.func = containCosineDistance
        elif method == "containJaccardDistance":
            self.func = containJaccardDistance
        elif method == "containDiceDistance":
            self.func = containDiceDistance
        elif method == "embedDistance":
            self.func = embedDistance
            self.embedding = spacy.load("en_core_web_lg")
        else:
            raise Exception("{} is an invalid distance function"
                             .format(method))

    def compute_distance(self, LR, weight=None):
        """"Compute distance score between tuple pairs

        Parameters:
        ----------
        LR: pd.DataFrame
            A table of tuple pairs. The columns of left and right values are
            named as "value_l" and "value_r". For char-based distance the type
            of values are string. For set-based distance the type of values are
            token set.

        weight: dict, default=None
            Weighting schema. If none, uniform weight or no weight is used.

        Return:
        -------
        distance: pd.Series
            distance between tuple pairs for each row
        """
        if weight is None:
            if self.method != "embedDistance":
                distance = LR.apply(lambda x: self.func(x.value_l, x.value_r), axis=1)
            else:
                distance = LR.apply(lambda x: self.func(x.value_l, x.value_r, self.embedding), axis=1)
        else:
            distance = LR.apply(lambda x: self.func(x.value_l, x.value_r, weight), axis=1)
        return distance

# data = pd.read_csv("../../data/left.csv")["title"]
# X = np.concatenate([data.values for _ in range(20)])
# X = pd.Series(X)
#
# L = X
# R = X.sample(frac=1)
#
# from tokenizer import Tokenizer
# tokenizer = Tokenizer("splitBySpace")
# L = tokenizer.tokenize(L)
# R = tokenizer.tokenize(R)
# LR = pd.DataFrame({"value_l":L, "value_r":R})
#
# tic = time.time()
# methods = ["jaccardDistance", "maxincDistance", "containCosineDistance"]
# distance_function = DistanceFunction("jaccardDistance")
# distance_function.compute_distance(LR)
# distance_function = DistanceFunction("maxincDistance")
# distance_function.compute_distance(LR)
# distance_function = DistanceFunction("containCosineDistance")
# distance_function.compute_distance(LR)
# print(time.time() - tic)



"""Options of join functions"""

autofj_lg = {
    "preprocess_methods":["lower", "lowerStem", "lowerRemovePunctuation",
                          "lowerRemovePunctuationStem"],
    "tokenize_methods": ["threeGram", "splitBySpace"],
    "token_weights": ["uniformWeight", "idfWeight"],
    "char_distance_functions": ["editDistance", "jaroDistance"], 
    "set_distance_functions": ["containJaccardDistance",
                               "containCosineDistance",
                               "containDiceDistance",
                               "intersectDistance",
                               "jaccardDistance",
                               "cosineDistance",
                               "diceDistance",
                               "maxincDistance"]
}

autofj_md = {
    "preprocess_methods":["lower", "lowerRemovePunctuationStem"],
    "tokenize_methods": ["threeGram", "splitBySpace"],
    "token_weights": ["uniformWeight", "idfWeight"],
    "char_distance_functions": ["editDistance", "jaroDistance"],
    "set_distance_functions": ["containJaccardDistance",
                               "containCosineDistance",
                               "containDiceDistance",
                               "intersectDistance",
                               "jaccardDistance",
                               "cosineDistance",
                               "diceDistance",
                               "maxincDistance"]
}

autofj_sm = {
    "preprocess_methods":["lower", "lowerRemovePunctuationStem"],
    "tokenize_methods": ["threeGram", "splitBySpace"],
    "token_weights": ["idfWeight"],
    "char_distance_functions": ["jaroDistance"],
    "set_distance_functions": ["containCosineDistance",
                               "jaccardDistance",
                               "maxincDistance"]
}


import numpy as np
from nltk.stem.porter import PorterStemmer
from nltk.stem import SnowballStemmer
import time
import pandas as pd
import re

ps = PorterStemmer()
# ps = SnowballStemmer("english")

def lower(x):
    return str(x).lower()

def removePunctuation(x):
    return re.sub(r'[^\w\s]', '', x)

def stem(x):
    return " ".join([ps.stem(w) for w in x.split()])

def lowerStem(x):
    x = lower(x)
    x = stem(x)
    return x

def lowerRemovePunctuation(x):
    x = lower(x)
    x = removePunctuation(x)
    return x

def lowerRemovePunctuationStem(x):
    x = lower(x)
    x = removePunctuation(x)
    x = stem(x)
    return x

class Preprocessor:
    """Preprocess data

    Parameters
    ----------
    method: string
        Preprocessing method. The available methods are listed as follows.
        - lower: lowercase
        - lowerStem: lowercase and stem
        - lowerRemovePunctuation: lowercase and remove punctuation
        - lowerRemovePunctuationStem: lowercase, remove punctuation and stem
    """
    def __init__(self, method):
        self.method = method
        if method == "lower":
            self.func = lower
        elif method == "lowerStem":
            self.func = lowerStem
        elif method == "lowerRemovePunctuation":
            self.func = lowerRemovePunctuation
        elif method == "lowerRemovePunctuationStem":
            self.func = lowerRemovePunctuationStem
        else:
            raise Exception("{} is an invalid preprocessing method"
                             .format(method))
    def preprocess(self, X):
        """ Preprocess the given data

        Parameters
        ----------
        X: pd.Series
            Input data
        """
        X = X.apply(self.func)
        return X

# data = pd.read_csv("../../data/left.csv")["title"]
# X = np.concatenate([data.values for _ in range(20)])
# X = pd.Series(X)

# pre1 = Preprocessor("lowerRemovePunctuationStem")
# tic = time.time()
# pre1.preprocess(X)
# print(time.time() - tic)

# pre2 = OldPreprocess(X)
# tic = time.time()
# pre2.process(("lower", "remove_punctuation", "stem"))
# print(time.time() - tic)




from .preprocessor import Preprocessor
from .tokenizer import Tokenizer
from .token_weight import TokenWeight
from .distance_function import DistanceFunction
import pandas as pd
import os
import pickle
from ...utils import makedir


class AutoFJJoinFunction(object):
    """Join function. A join function computes a distance score between a
    pair of records on a given column. Our built-in join function is specified
    by four parameters including preprocess_method, tokenize_method,
    token_weight_method and distance_function.

    Parameters
    ----------
    preprocess_method: string
        Preprocessing method. See preprocessor.py for available methods.

    tokenize_method: string or None
        Tokenization method. See tokenizer.py for available methods.

    token_weight_method: string or None
        Token weighting method. See token_weight.py for available methods.

    distance_function: string
        Distance function. See distance_function.py for available methods.

    """

    def __init__(self,
                 preprocess_method,
                 tokenize_method,
                 token_weight_method,
                 distance_function):
        self.preprocess_method = preprocess_method
        self.tokenize_method = tokenize_method
        self.token_weight_method = token_weight_method
        self.distance_function = distance_function
        self.name = "{}_{}_{}_{}".format(preprocess_method,
                                         tokenize_method,
                                         token_weight_method,
                                         distance_function)

    def compute_distance(self, left, right, LL_blocked, LR_blocked,
                         cache_dir=None):
        """Compute the distance of each tuple pair in the LL and LR blocked table.

        Parameters
        ----------
        left: pd.DataFrame
            A subset of the left table that contains the id column and the
            column to be processed. The id column is named as autofj_id.
            The column to be processed is named as value.

        right: pd.DataFrame
            A subset of the right table that contains the id column and the
            column to be processed. The id column is named as autofj_id.
            The column to be processed is named as value.

        LL_blocked: pd.DataFrame
            The LL blocked table that consists of the id columns and
            the columns to be processed. The id columns are named as
            autofj_id_l and autofj_id_r. The column to be processed is named as
            value_l and value_r.

        LR_blocked: pd.DataFrame
            The LR blocked table that consists of the id columns and
            the columns to be processed. The id columns are named as
            autofj_id_l and autofj_id_r. The column to be processed is named as
            value_l and value_r.

        Returns
        -------
        LL_distance: pd.Series
            Distance of each tuple pair in the LL blocked table.

        LR_distance: pd.Series
            Distance of each tuple pair in the LR blocked table.
        """
        left = left.set_index("autofj_id")
        right = right.set_index("autofj_id")

        # cache paths
        if cache_dir is not None:
            L_pre_path, L_pre_token_path, token_weight_path\
                = self.get_cache_paths(cache_dir, "left")
            R_pre_path, R_pre_token_path, _\
                = self.get_cache_paths(cache_dir, "right")
        else:
            L_pre_path = None
            R_pre_path = None
            L_pre_token_path = None
            R_pre_token_path = None
            token_weight_path = None

        # preprocess
        L_pre = self.preprocess(left["value"], cache_path=L_pre_path)
        R_pre = self.preprocess(right["value"], cache_path=R_pre_path)

        # tokenize
        L_pre_token = self.tokenize(L_pre, cache_path=L_pre_token_path)
        R_pre_token = self.tokenize(R_pre, cache_path=R_pre_token_path)

        # token weights
        token_weights = self.compute_token_weights(L_pre_token,
                                                   cache_path=token_weight_path)

        # apply distance function
        LL_distance = self.apply_distance_function(LL_blocked,
                                                   L_pre_token,
                                                   L_pre_token,
                                                   token_weights)
        LR_distance = self.apply_distance_function(LR_blocked,
                                                   L_pre_token,
                                                   R_pre_token,
                                                   token_weights)
        return LL_distance, LR_distance

    def get_cache_paths(self, cache_dir, table):
        """Get paths of different cache files

        Parameters
        ----------
        cache_dir: string
            Directory to store cache

        table: string
            Either "left" or "right"

        Returns
        -------
        pre_path:
            Path of preprocessing results

        pre_token_path:
            Path of preprocessing and tokenization results

        token_weight_path:
            Path of token weights
        """
        pre_path = makedir([cache_dir, table],
                             "{}.p".format(self.preprocess_method))

        pre_token_path = makedir([cache_dir, table],
                            "{}_{}.p".format(self.preprocess_method,
                                             self.tokenize_method))

        if table == "left":
            token_weight_path = makedir([cache_dir, table],
                                        "{}_{}_{}.p".format(
                                            self.preprocess_method,
                                            self.tokenize_method,
                                            self.token_weight_method))
        else:
            token_weight_path = None
        return pre_path, pre_token_path, token_weight_path

    def preprocess(self, X, cache_path=None):
        """Preprocess records

        Parameters
        ----------
        X: pd.Series
            Input data

        Returns
        -------
        X_pre: pd.Series
            Preprocessed data.
        """
        if cache_path is not None and os.path.exists(cache_path):
            with open(cache_path, "rb") as f:
                X_pre = pickle.load(f)
        else:
            preprocessor = Preprocessor(self.preprocess_method)
            X_pre = preprocessor.preprocess(X)
            self.save_cache(X_pre, cache_path)
        return X_pre

    def tokenize(self, X_pre, cache_path=None):
        """ Tokenize the preprocessed data

        Parameters
        ----------
        X_pre: pd.Series
            Preprocessed data

        Returns
        -------
        X_pre_token: pd.Series
            Data after preprocessing and tokenization
        """
        if cache_path is not None and os.path.exists(cache_path):
            with open(cache_path, "rb") as f:
                X_pre_token = pickle.load(f)
        else:
            tokenizer = Tokenizer(self.tokenize_method)
            X_pre_token = tokenizer.tokenize(X_pre)
            self.save_cache(X_pre_token, cache_path)
        return X_pre_token

    def compute_token_weights(self, X_pre_token, cache_path=None):
        """ Compute token weights

        Parameters
        ----------
        X_pre_token: pd.Series
            Data after preprocessing and tokenization

        Returns
        -------
        token_weights: dict
            Token weights
        """
        if cache_path is not None and os.path.exists(cache_path):
            with open(cache_path, "rb") as f:
                token_weights = pickle.load(f)
        else:
            weight = TokenWeight(self.token_weight_method)
            token_weights = weight.weight(X_pre_token)
            self.save_cache(token_weights, cache_path)
        return token_weights

    def apply_distance_function(self, LR_blocked, L_pre_token, R_pre_token,
                                token_weights):
        """ Apply distance functions on blocked record pairs

        Parameters
        ----------
        LR_blocked: pd.DataFrame
            A subset of the blocked table that contains the id columns and
            the columns to be processed. The id columns are named as
            autofj_id_l and autofj_id_r. The column to be processed is named as
            value_l and value_r.

        L_pre_token: pd.Series
            Left records after preprocessing and tokenization

        R_pre_token: pd.Series
            Right records after preprocessing and tokenization

        token_weights: dict
            Token weights

        Returns
        -------
        distance: pd.Series
            distance between record pairs in LR_blocked
        """
        # apply join function on record pairs in LR_blocked
        L_id = LR_blocked["autofj_id_l"].values
        R_id = LR_blocked["autofj_id_r"].values

        L = L_pre_token.loc[L_id]
        R = R_pre_token.loc[R_id]
        LR = pd.DataFrame({"value_l": L.values, "value_r": R.values})

        df = DistanceFunction(self.distance_function)
        distance = df.compute_distance(LR, token_weights).values
        return distance

    def save_cache(self, data, cache_path):
        if cache_path is not None:
            with open(cache_path, "wb") as f:
                pickle.dump(data, f)

# if __name__ == '__main__':
#
#     left = pd.read_csv("../../data/left.csv")[["id", "name"]]
#     left = left.rename(columns={"id":"autofj_id", "name":"value"})
#     right = pd.read_csv("../../data/left.csv")[["id", "name"]]
#     right = right.rename(columns={"id":"autofj_id", "name":"value"})
#     LR_blocked = pd.read_csv("../../data/LR_blocked.csv")
#
#     jf = AutoFJJoinFunction("lower", None, None, "editDistance")
#     tic = time.time()
#     distance = jf.compute_distance(left, right, LR_blocked)
#     print(time.time() - tic)



import pandas as pd
import os
from os.path import dirname

def load_data(name):
    module_path = dirname(__file__)
    if os.path.exists(os.path.join(module_path, "benchmark", name)):
        left_table = pd.read_csv(os.path.join(module_path, "benchmark", name, "left.csv"))
        right_table = pd.read_csv(os.path.join(module_path, "benchmark", name, "right.csv"))
        gt_table = pd.read_csv(os.path.join(module_path, "benchmark", name, "gt.csv"))
        return left_table, right_table, gt_table
    else:
        print(os.path.join(module_path, "benchmark", name))
        raise Exception("Dataset {} does not exist.".format(name))



import collections
import math
import pandas as pd
import time
import numpy as np

class defaultdict(dict):
    def set_default_value(self, default_value):
        self.default_value = default_value
        
    def __missing__(self, key):
        self[key] = self.default_value
        return self.default_value

def uniformWeight(document):
    """Uniform weight"""
    weight = defaultdict()
    weight.set_default_value(1)
    return weight

def idfWeight(document):
    """Compute idf weight for tokens

    Parameters:
    -----------
    document: list of sets
        A list of token sets, which is the document on which the idf is
        computed.

    Return:
    -------
    weight: dict
        idf weight of tokens
    """
    token_count = collections.defaultdict(set)

    for i, row in enumerate(document):
        for token in row:
            token_count[token].add(i)

    # calculate idf value
    weight = defaultdict()
    weight.set_default_value(math.log(len(document)))

    for k, v in token_count.items():
        weight[k] = math.log(len(document) / (len(v) + 1))
    return weight

class TokenWeight(object):
    """Token weight

    Parameters
    ----------
    method: string
        Token weighting schema. The available methods are listed as follows.
        - uniformWight
        - idfWeight
        - None (no weights)
    """
    def __init__(self, method):
        self.method = method
        if method is None:
            self.func = None
        elif method == "uniformWeight":
            self.func = uniformWeight
        elif method == "idfWeight":
            self.func = idfWeight
        else:
            raise Exception("{} is an invalid weighting schema"
                             .format(method))
    
    def weight(self, X):
        """ Weight tokens

        Parameters
        ----------
        X: pd.Series
            Input data

        Return
        ------
        weight: dict
            weight of tokens
        """
        if self.func is not None:
            weight = self.func(X)
            return weight
        else:
            return None
#
# data = pd.read_csv("../../data/left.csv")["title"]
# X = np.concatenate([data.values for _ in range(20)])
# X = pd.Series(X)
#
# weight = TokenWeight("idfWeight")
# tic = time.time()
# weight.weight(X)
# print(time.time() - tic)



import pandas as pd
import numpy as np
from multiprocessing import Pool
from .autofj_single_column_greedy_algorithm import AutoFJGreedyAlgorithm
from ..utils import print_log
import os

class AutoFJMulticolGreedyAlgorithm(object):
    """Greedy algorithm for multi-column datasets. Select optimal columns and
    column weights using column forward selection.

    Parameters
    ----------
    LL_distance: dict
        Distance of tuple pairs in LL tables measured by different join
        functions. The distance is saved in a dict. The key is the name
        of join functions. The value is a table of distance between
        tuple pairs on different columns. The first two columns in the
        table are "autofj_id_l", "autofj_id_r". The remaining columns
        are distance on different columns.

    LR_distance: dict
        Distance of tuple pairs in LR tables measured by different join
        functions. The distance is saved in a dict. The key is the name
        of join functions. The value is a pd.DataFrame of distance between
        tuple pairs on different columns. The first two columns in the
        table are "autofj_id_l", "autofj_id_r". The remaining columns
        are distance on different columns.

    precision_target: float
        Precision target. This should be a float number between 0-1.

    candidate_thresholds: list
        The search space of distance threshold.

    candidate_column_weights: list
        The search space of column weights.

    n_jobs : int, default=1
        Number of CPU cores used. -1 means using all processors.

    verbose: bool, default=False
        Whether to print logging

    """
    def __init__(self,
                 LL_distance,
                 LR_distance,
                 precision_target,
                 candidate_thresholds,
                 candidate_column_weights,
                 n_jobs=-1,
                 verbose=False):
        self.precision_target = precision_target
        self.LL_distance = LL_distance
        self.LR_distance = LR_distance
        self.join_functions = sorted(list(LL_distance.keys()))
        self.candidate_thresholds = candidate_thresholds
        self.candidate_column_weights = candidate_column_weights
        self.verbose = verbose
        self.column_names = []
        for c in LL_distance[self.join_functions[0]].columns:
            if c != "autofj_id_l" and c != "autofj_id_r":
                self.column_names.append(c)
        self.n_jobs = n_jobs if n_jobs > 0 else os.cpu_count()

    def run(self):
        """Running forward selection algorithm"""
        best_reward = float("-inf")
        best_LR_joins = None
        best_columns = []
        best_weights = None
        best_join_config = None
        best_column_weights = None

        for i in range(len(self.column_names)):
            # get the best result after adding one column
            columns, weights, join_config, LR_joins, reward \
                = self.forward_selection(best_columns, best_weights)

            # if the reward stops increasing by adding columns, terminates
            if reward <= best_reward:
                break

            # save best result
            best_columns, best_weights, best_join_config, best_LR_joins, best_reward \
                = columns, weights, join_config, LR_joins, reward

            if best_join_config is not None:
                best_column_weights = self.get_column_weights(best_columns,
                                                              best_weights)
            if self.verbose:
                print_log("Best column_weight: {}:{}, Best reward: {}"
                      .format(",".join(best_columns),
                              ",".join([str(w) for w in best_weights]),
                              best_reward))

        return best_column_weights, best_join_config, best_LR_joins

    def forward_selection(self, base_columns, base_weights):
        """Do one step forward selection. Adding one column from the remaining
        columns, get the best column and weights.

        Parameters
        ----------
        base_columns: list
            Old columns (best column from the last iteration)

        base_weights: list
            Old weights (best weight from the last iteration)

        Returns
        -------
        best_columns: list
            Best columns after adding one column

        best_weights: list
            Best weights after adding one column

        best_pred: list
            LR prediction

        best_reward: float
            resulting reward (profit or estimated recall) given the best
            columns and weights
        """
        # get all candidate column weights
        column_weights_cands = self.get_column_weights_cands(base_columns,
                                                             base_weights)

        if self.n_jobs == 1:
            results = []
            for column_weights in column_weights_cands:
                res = self.run_one_weight(column_weights)
                results.append(res)
        else:
            with Pool(self.n_jobs) as pool:
                results = pool.map(self.run_one_weight, column_weights_cands)

        best_reward = float("-inf")
        best_weights = None
        best_LR_joins = None
        best_join_config = None
        best_columns = None

        for i, (LR_joins, reward, config_selected) in enumerate(results):
            if reward > best_reward:
                best_reward = reward
                best_columns = list(column_weights_cands[i].keys())
                best_weights = list(column_weights_cands[i].values())
                best_LR_joins = LR_joins
                best_join_config = config_selected
        return best_columns, best_weights, best_join_config, best_LR_joins, \
               best_reward

    def get_column_weights_cands(self, base_columns, base_weights):
        """Get candidate column weights by adding one column into old columns

        Parameters
        ----------
        base_columns: list
            Old columns (best column from the last iteration)

        base_weights: list
            Old weights (best weight from the last iteration)

        Returns
        -------
        column_weights_cands: list
            A list of candidate column weights (dict)

        """
        """ get column-weight candidates"""
        column_weights_cands = []
        for c in self.column_names:
            if c in base_columns:
                continue
            columns = base_columns + [c]

            # get search space (weights candidates) for new weights
            new_weights = self.get_new_weights(base_weights)

            for weights in new_weights:
                column_weights = self.get_column_weights(columns, weights)
                column_weights_cands.append(column_weights)
        return column_weights_cands

    def run_one_weight(self, column_weights):
        # if self.verbose:
        #     print_log("Run greedy algorithm with column weights"
        #           .format(column_weights)

        LL_w = self.get_weighted_distance(self.LL_distance, column_weights)
        LR_w = self.get_weighted_distance(self.LR_distance, column_weights)
        optimizer = AutoFJGreedyAlgorithm(LL_w,
                                          LR_w,
                                          self.precision_target,
                                          self.candidate_thresholds,
                                          n_jobs=self.n_jobs)
        LR_joins, config_selected = optimizer.run()
        reward = optimizer.get_reward()
        return LR_joins, reward, config_selected

    def get_new_weights(self, old_weights):
        """Get column weight search space. Keeping ratios of old weights fixed,
           append weight for the new column.

        Parameters
        ----------
        old_weights: list
             Weights of old columns

        Return
        ------
        column_weights: list
            A list of new weights. In each new weight, the last one is the new
            weight, the others are old weights. The ratio between old weights
            is fixed.
            Example: old_weights: [0.5, 0.5]
                column_weights = [[0.45, 0.45, 0.1],
                                  [0.40, 0.40, 0.2],
                                  [0.35, 0.35, 0.3],
                                  ...]
        """
        # if old weight is empty, new weight is [1].
        if old_weights is None:
            return np.array([[1]])

        # add weight for new column
        column_weights = []
        for nw in self.candidate_column_weights[1:-1]:
            new_w = np.array([w * nw for w in old_weights] + [1 - nw])
            column_weights.append(new_w)

        return column_weights

    def get_weighted_distance(self, LL_distance, column_weights):
        """LL_w: {config: [lid, rid, distance]}"""
        LL_w = {}
        for config, distance in LL_distance.items():
            columns = [c for c in distance.columns if
                       c not in ["autofj_id_l", "autofj_id_r"]]
            weights = []
            for c in columns:
                if c in column_weights:
                    weights.append(column_weights[c])
                else:
                    weights.append(0)
            weights = np.array(weights).reshape(-1, 1)
            weighted_dist = distance[columns].values.dot(weights).ravel()
            LL_w[config] = pd.DataFrame(
                {"autofj_id_l": distance["autofj_id_l"],
                 "autofj_id_r": distance["autofj_id_r"],
                 "distance": weighted_dist})
        return LL_w

    def get_column_weights(self, columns, weights):
        column_weights = {}
        for c, w in zip(columns, weights):
            column_weights[c] = w
        return column_weights